from enum import Enum, auto


class Tokens(Enum):
    WhiteSpace = 'WhiteSpace'
    Space = ' '
    Tab = '\t'
    Enter = '\n'

    SingleLineComment = '//'
    MultiLineComment = '/*'

    Keyword = ('__halt_compiler', 'abstract', 'and', 'array', 'as', 'break', 'callable', 'case', 'catch', 'class',
               'clone', 'const', 'continue', 'declare', 'default', 'die', 'do', 'echo', 'else', 'elseif', 'empty',
               'enddeclare', 'endfor', 'endforeach', 'endif', 'endswitch', 'endwhile', 'eval', 'exit', 'extends',
               'final', 'finally', 'fn', 'for', 'foreach', 'function', 'global', 'goto', 'if', 'implements', 'include',
               'include_once', 'instanceof', 'insteadof', 'interface', 'isset', 'list',	'namespace', 'new', 'or',
               'print', 'private', 'protected', 'public', 'require', 'require_once', 'return', 'static', 'switch',
               'throw', 'trait', 'try', 'unset', 'use', 'var', 'while', 'xor', 'yield', 'from',

               'int', 'float', 'bool', 'string', 'void', 'iterable', 'object', 'resource', 'mixed', 'numeric',

               'Directory', 'stdClass', '__PHP_Incomplete_Class', 'Exception', 'ErrorException', 'php_user_filter',
               'Closure', 'Generator', 'ArithmeticError', 'AssertionError', 'DivisionByZeroError', 'Error',
               'Throwable', 'ParseError', 'TypeError', 'self', 'static', 'parent')

    ConstLiteral = ('null', 'true', 'false',

                    '__CLASS__', '__DIR__', '__FILE__', '__FUNCTION__', '__LINE__',
                    '__METHOD__', '__NAMESPACE__', '__TRAIT__',

                    'PHP_VERSION', 'PHP_MAJOR_VERSION', 'PHP_MINOR_VERSION', 'PHP_RELEASE_VERSION', 'PHP_VERSION_ID',
                    'PHP_EXTRA_VERSION', 'PHP_ZTS', 'PHP_DEBUG', 'PHP_MAXPATHLEN', 'PHP_OS', 'PHP_OS_FAMILY',
                    'PHP_SAPI', 'PHP_EOL', 'PHP_INT_MAX', 'PHP_INT_MIN', 'PHP_INT_SIZE', 'PHP_FLOAT_DIG',
                    'PHP_FLOAT_EPSILON', 'PHP_FLOAT_MIN', 'PHP_FLOAT_MAX', 'DEFAULT_INCLUDE_PATH', 'PEAR_INSTALL_DIR',
                    'PEAR_EXTENSION_DIR', 'PHP_EXTENSION_DIR', 'PHP_PREFIX', 'PHP_BINDIR', 'PHP_BINARY', 'PHP_MANDIR',
                    'PHP_LIBDIR', 'PHP_DATADIR', 'PHP_SYSCONFDIR', 'PHP_LOCALSTATEDIR', 'PHP_CONFIG_FILE_PATH',
                    'PHP_CONFIG_FILE_SCAN_DIR', 'PHP_SHLIB_SUFFIX', 'PHP_FD_SETSIZE', 'E_ERROR', 'E_WARNING', 'E_PARSE',
                    'E_NOTICE', 'E_CORE_ERROR', 'E_CORE_WARNING', 'E_COMPILE_ERROR', 'E_COMPILE_WARNING',
                    'E_USER_ERROR', 'E_USER_WARNING', 'E_USER_NOTICE', 'E_RECOVERABLE_ERROR', 'E_DEPRECATED',
                    'E_USER_DEPRECATED', 'E_ALL', 'E_STRICT', '__COMPILER_HALT_OFFSET__', 'TRUE', 'FALSE', 'NULL',
                    'PHP_WINDOWS_EVENT_CTRL_C', 'PHP_WINDOWS_EVENT_CTRL_BREAK')
    NumberLiteral = ('decimal_float',
                     'decimal_int', 'binary', 'octal', 'hexadecimal',
                     'B_decimal_int', 'B_binary', 'B_octal', 'B_hexadecimal')
    Identifier = 'identifier'
    Variable = 'variable'
    StringLiteral = ('\'', '"', '`')
    StartTemplateString = 'start "'
    EndTemplateString = 'end "'
    InterpolationStart = '${'
    InterpolationEnd = '}'

    Punctuation = ('[', ']', '(', ')', '{', '}', ';', ',')

    Operators = ('=', '+=', '-=', '*=', '/=', '%=', '**=', '&=', '|=', '^=', '<<=', '>>=', '??=',
                 '&&', '||', '!',
                 '==', '===', '!=', '<>', '!==', '>', '<', '>=', '<=', '<=>',
                 '&', '|', '^', '~',
                 '+', '-',
                 '*', '/', '%', '**',
                 '<<', '>>',
                 '++', '--',
                 '?', ':',
                 '.', '.=',
                 '->',
                 '<?',
                 '?>',
                 '\\'
                 )
    Regex = 'regex'
    Invalid = 'Invalid'
    Interpolation = 'Interpolation'


from php_formatter import PHPFormatter
import sys
import os
from os import listdir
from os.path import isfile, join, splitext
import logging


def _get_dirs_from_path(my_path):
    return [f for f in listdir(my_path) if not isfile(join(my_path, f))]


def _get_php_files_from_path(my_path):
    all_files = [f for f in listdir(my_path) if isfile(join(my_path, f))]
    js_files = []
    for f in all_files:
        f_name, f_ext = splitext(f)
        if f_ext == '.php':
            js_files.append(f)
    return js_files


def _get_result(file_path, file_name):
    formatter = PHPFormatter()
    formatter.process_php_file(file_path, file_name)
    PHPFormatter.formatted_files.append(formatter)
    PHPFormatter.all_changes.extend(formatter.get_invalid_tokens_list())
    PHPFormatter.old_file_names.append(formatter.file_name)
    PHPFormatter.new_file_names.append(formatter.new_file_name)


def _check_files_in_dir(my_path):
    php_files = _get_php_files_from_path(my_path)
    for php_file in php_files:
        _get_result(my_path, php_file)


def _check_rec(my_path):
    _check_files_in_dir(my_path)

    dirs = _get_dirs_from_path(my_path)
    for dir in dirs:
        _check_rec(join(my_path, dir))


def main():
    if len(sys.argv) == 2 and (sys.argv[1] == '-h' or sys.argv[1] == '--help'):
        print("Basic commands:")
        print("\t-h, --help\t\t\t\t: help menu")
        print("\t-v, --verify -(p|d|f) /..\t: verify your files(as output log.txt file);")
        print("\t\t\t\t\t\t  /.. - path to project, directory or file")

        print("\t-f, --format -(p|d|f) /..\t: format your files(as output log.txt file);")
        print("\t\t\t\t\t\t  /.. - path to project, directory or file")
    elif len(sys.argv) == 4 and sys.argv[2] == '-p':
        _check_rec(sys.argv[3])
    elif len(sys.argv) == 4 and sys.argv[2] == '-d':
        _check_files_in_dir(sys.argv[3])
    elif len(sys.argv) == 4 and sys.argv[2] == '-f':
        _get_result('', sys.argv[3])
    else:
        print("Call help menu (-h, --help) for more details")
        exit()

    if sys.argv[1] == '-f' or sys.argv[1] == '--format':
        with open("php_fixing.log", "w") as f:
            pass
        logging.basicConfig(filename="php_fixing.log", level=logging.DEBUG)
        for file in PHPFormatter.formatted_files:
            file.setup_all_changes()
            file.save_to_file()
            file.save_fix_log()
    elif sys.argv[1] == '-v' or sys.argv[1] == '--verify':
        with open("php_verification.log", "w") as f:
            pass
        logging.basicConfig(filename="php_verification.log", level=logging.DEBUG)
        for file in PHPFormatter.formatted_files:
            file.setup_all_changes()
            file.save_ver_log()



def snake_case(s):
    new_s = ''
    for i in range(len(s)):
        if (s[i]).isupper() and i != 0 and ((s[i - 1]).islower() or i + 1 < len(s) and (s[i + 1]).islower()):
            new_s += '_'
        new_s += (s[i]).lower()
    return new_s


def screaming_snake_case(s):
    return s.upper()


def camel_case(s):
    new_s = ''
    is_first = True
    for i in range(len(s)):
        if s[i] != '_':
            if is_first:
                new_s += (s[i]).upper()
                is_first = False
            elif i > 0 and s[i-1] == '_':
                new_s += s[i].upper()
            else:
                new_s += s[i]

    return new_s


def file_name_case(s):
    new_s = ''
    for i in range(len(s)):
        if s[i] != '_':
            if (s[i]).isupper() and i != 0 and ((s[i - 1]).islower() or i + 1 < len(s) and (s[i + 1]).islower()):
                new_s += '-'
            new_s += (s[i]).lower()
        elif i > 0 and s[i-1] not in '-_':
            new_s += '-'
    return new_s



import sys
import os
import logging
from os.path import isfile, join, splitext

from ..lexer import Lexer
from ..lexer.token_classes import Token, WrongToken
from ..lexer.dict_token_types import Tokens

from ._cases_creation import *


class PHPFormatter:
    formatted_files = []
    all_changes = []
    old_file_names = []
    new_file_names = []

    def __init__(self):
        self.path_to_file = ''
        self.file_name = ''
        self.new_file_name = ''

        self._token = []
        self._invalid_token = []
        self._includes = []
        self._symbol_table = []
        self._lexer_error_list = []

        self._state_pos = 0
        self._all_tokens = []

    def process_php_file(self, path_to_file, file_name):
        file_path = join(path_to_file, file_name)
        self.path_to_file = path_to_file
        self.file_name = file_name
        self.new_file_name = file_name_case(file_name)

        lexer = Lexer()
        lexer.process_php_file(file_path)
        self._all_tokens = lexer.get_tokens_list()
        self._symbol_table = lexer.get_symbol_table()
        self._lexer_error_list = lexer.get_error_tokens_list()

        while self._state_pos < len(self._all_tokens):
            self._parse_next_token()

    def _parse_next_token(self):
        current_token = self._all_tokens[self._state_pos]

        if current_token.type == Tokens.Variable:
            self._handle_variable()
            return
        if current_token.spec == 'function':
            self._handle_function()
            return
        if current_token.spec == 'class':
            self._handle_class()
            return
        if current_token.spec == 'namespace':
            self._handle_namespace()
            return

        if current_token.spec == 'include':
            self._handle_include()
            return

        self._token.append(self._all_tokens[self._state_pos])
        self._state_pos += 1

    def _is_whitespace_token(self, token):
        return token == Tokens.Enter or token == Tokens.Space or token == Tokens.Tab

    def _add_new_token(self, current_token, new_value, error_message, format_type):
        new_token = Token(
            token_type=current_token.type,
            row=current_token.row,
            column=current_token.column,
            index=len(self._symbol_table)
        )

        self._token.append(new_token)
        self._symbol_table.append(new_value)
        self._invalid_token.append(WrongToken(
            token=current_token,
            message=error_message,
            format_type=format_type,
            new_value=new_value,
            old_value=self._symbol_table[current_token.index]
        ))

    def _handle_variable(self):
        normal_end = True
        while self._next_non_whitespace(self._state_pos + 1).spec == '->':
            current_token = self._all_tokens[self._state_pos]
            variable = self._symbol_table[current_token.index]
            if current_token.type == Tokens.Variable:
                new_variable_lower = '$' + snake_case(variable[1:])
                new_variable_upper = '$' + screaming_snake_case(variable[1:])
            else:
                new_variable_lower = snake_case(variable)
                new_variable_upper = screaming_snake_case(variable)
            if new_variable_upper == variable:
                self._token.append(current_token)
            elif new_variable_lower != variable:
                self._add_new_token(current_token, new_variable_lower, "Incorrect snake case in variable name", "var")
            else:
                self._token.append(current_token)
            self._state_pos += 1
            while self._state_pos < len(self._all_tokens) and\
                    self._is_whitespace_token(self._all_tokens[self._state_pos].type):
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1

            if self._state_pos >= len(self._all_tokens):
                return

            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1
            next_token = self._next_non_whitespace(self._state_pos)
            if next_token.type != Tokens.Identifier:
                normal_end = False
                break
            while self._state_pos < len(self._all_tokens) and\
                    self._is_whitespace_token(self._all_tokens[self._state_pos].type):
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1

        if not normal_end or self._next_non_whitespace(self._state_pos + 1).spec == '(' and\
                self._all_tokens[self._state_pos].type != Tokens.Variable:
            return
        current_token = self._all_tokens[self._state_pos]
        variable = self._symbol_table[current_token.index]
        if current_token.type == Tokens.Variable:
            new_variable_lower = '$' + snake_case(variable[1:])
            new_variable_upper = '$' + screaming_snake_case(variable[1:])
        else:
            new_variable_lower = snake_case(variable)
            new_variable_upper = screaming_snake_case(variable)
        if new_variable_upper == variable:
            self._token.append(current_token)
        elif new_variable_lower != variable:
            self._add_new_token(current_token, new_variable_lower, "Incorrect snake case in variable name", "var")
        else:
            self._token.append(current_token)
        self._state_pos += 1

    def _handle_function(self):
        next_token = self._next_non_whitespace(self._state_pos + 1)
        self._token.append(self._all_tokens[self._state_pos])
        self._state_pos += 1

        if next_token.type != Tokens.Identifier:
            return

        while self._state_pos < len(self._all_tokens) and \
                self._is_whitespace_token(self._all_tokens[self._state_pos].type):
            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1

        current_token = self._all_tokens[self._state_pos]
        func_name = self._symbol_table[current_token.index]
        new_func_name = snake_case(func_name)
        if func_name != new_func_name:
            self._add_new_token(current_token, new_func_name, "Incorrect snake case in func name", "func")
        else:
            self._token.append(current_token)
        self._state_pos += 1

    def _handle_class(self):
        next_token = self._next_non_whitespace(self._state_pos + 1)
        self._token.append(self._all_tokens[self._state_pos])
        self._state_pos += 1

        if next_token.type != Tokens.Identifier:
            return

        while self._state_pos < len(self._all_tokens) and \
                self._is_whitespace_token(self._all_tokens[self._state_pos].type):
            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1

        current_token = self._all_tokens[self._state_pos]
        class_name = self._symbol_table[current_token.index]
        new_class_name = camel_case(class_name)
        if class_name != new_class_name:
            self._add_new_token(current_token, new_class_name, "Incorrect snake case in class name", "class")
        else:
            self._token.append(current_token)
        self._state_pos += 1

    def _handle_namespace(self):
        next_token = self._next_non_whitespace(self._state_pos + 1)
        self._token.append(self._all_tokens[self._state_pos])
        self._state_pos += 1

        if next_token.type != Tokens.Identifier:
            return
        while self._state_pos < len(self._all_tokens) and \
                self._is_whitespace_token(self._all_tokens[self._state_pos].type):
            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1

        normal_end = True

        while self._next_non_whitespace(self._state_pos + 1).spec == '\\':
            current_token = self._all_tokens[self._state_pos]
            namespace = self._symbol_table[current_token.index]
            new_namespace = camel_case(namespace)
            if new_namespace == namespace:
                self._token.append(current_token)
            elif new_namespace != namespace:
                self._add_new_token(current_token, new_namespace, "Incorrect snake case in namespace name", "namespace")
            else:
                self._token.append(current_token)
            self._state_pos += 1
            while self._state_pos < len(self._all_tokens) and \
                    self._is_whitespace_token(self._all_tokens[self._state_pos].type):
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1

            if self._state_pos >= len(self._all_tokens):
                return

            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1
            next_token = self._next_non_whitespace(self._state_pos)
            if next_token.type != Tokens.Identifier:
                normal_end = False
                break
            while self._state_pos < len(self._all_tokens) and \
                    self._is_whitespace_token(self._all_tokens[self._state_pos].type):
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1

        if not normal_end or self._state_pos >= len(self._all_tokens):
            return
        current_token = self._all_tokens[self._state_pos]
        namespace = self._symbol_table[current_token.index]
        new_namespace = camel_case(namespace)
        if new_namespace != namespace:
            self._add_new_token(current_token, new_namespace, "Incorrect snake case in namespace name", "namespace")
        else:
            self._token.append(current_token)
        self._state_pos += 1

    def _handle_include(self):
        next_token = self._next_non_whitespace(self._state_pos + 1)
        self._token.append(self._all_tokens[self._state_pos])
        self._state_pos += 1

        if not next_token.type == Tokens.StringLiteral and next_token == Tokens.StartTemplateString:
            return

        while self._state_pos < len(self._all_tokens) and \
                self._is_whitespace_token(self._all_tokens[self._state_pos].type):
            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1

        current_token = self._all_tokens[self._state_pos]
        if current_token.type == Tokens.StringLiteral:
            self._includes.append((current_token, len(self._token)))
            self._token.append(current_token)
            self._state_pos += 1
        else:
            pos = self._state_pos
            while pos < len(self._all_tokens) and self._all_tokens[pos].type != Tokens.EndTemplateString:
                pos += 1
            if pos - self._state_pos != 2:
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1
                return

            self._includes.append((self._all_tokens[self._state_pos + 1], len(self._token) + 1))
            while self._state_pos < len(self._all_tokens) and \
                    self._all_tokens[self._state_pos].type != Tokens.EndTemplateString:
                self._token.append(self._all_tokens[self._state_pos])
                self._state_pos += 1

            self._token.append(self._all_tokens[self._state_pos])
            self._state_pos += 1

    def _next_non_whitespace(self, pos):
        while pos < len(self._all_tokens) and self._is_whitespace_token(self._all_tokens[pos].type):
            pos += 1
        if pos >= len(self._all_tokens):
            return Token().set_invalid()
        else:
            return self._all_tokens[pos]

    def print_all(self):
        for tok in self._token:
            if tok.type != Tokens.Space and tok.type != Tokens.Tab:
                print(tok, end=' ')
                if tok.index is not None:
                    print('|' + self._symbol_table[tok.index] + '|')
                else:
                    print('')
            if tok.type == Tokens.Enter:
                print('')
        print('------------------------')
        for tok in self._invalid_token:
            print(tok, end=' ')
            if tok.token.index is not None:
                print('|' + self._symbol_table[tok.token.index] + '|')
            else:
                print('')

        print('------------------------')
        for i in range(len(self._symbol_table)):
            print(str(i) + ') ', '|' + self._symbol_table[i] + '|')

        print('------------------------')
        print(len(self._includes))
        for tok in self._includes:
            if tok[0].index is not None:
                print('|' + self._symbol_table[tok[0].index] + '|')
            else:
                print('')

    def get_tokens_list(self):
        return self._token

    def get_symbol_table(self):
        return self._symbol_table

    def get_includes(self):
        return self._includes

    def get_invalid_tokens_list(self):
        return self._invalid_token

    def save_to_file(self):
        os.remove(join(self.path_to_file, self.file_name))

        with open(join(self.path_to_file, self.new_file_name), 'w') as f:
            original_stdout = sys.stdout
            sys.stdout = f
            for token in self._token:
                if token.type == Tokens.StartTemplateString or token.type == Tokens.EndTemplateString:
                    print('"', end='')
                elif token.index is not None:
                    print(self._symbol_table[token.index], end='')
                elif token.spec is not None:
                    print(token.spec, end='')
                else:
                    print(token.type.value, end='')
            sys.stdout = original_stdout

    def save_ver_log(self):
        if self.file_name != self.new_file_name:
            logging.info('----------- ' + join(self.path_to_file, self.file_name) +
                         ' need to rename to ' + join(self.path_to_file, self.new_file_name) + ' -------------')
        else:
            logging.info('----------- ' + join(self.path_to_file, self.file_name) + ' -------------')
        logging.info('Verification errors:')
        i = 0
        for tok in self._invalid_token:
            if tok.token.index is not None:
                logging.info('Id: ' + str(i) + ') ' + join(self.path_to_file, self.file_name) + ': ' +
                             str(tok.token.row) + '|' + str(tok.token.column) + ' - Error code ( ' +
                             str(tok.old_value) + ' ): ' + str(tok.error_message))
            i += 1
        logging.info('errors end\n')

    def save_fix_log(self):
        if self.file_name != self.new_file_name:
            logging.info('----------- ' + join(self.path_to_file, self.file_name) +
                         ' renamed to ' + join(self.path_to_file, self.new_file_name) + ' -------------')
        else:
            logging.info('----------- ' + join(self.path_to_file, self.file_name) + ' -------------')
        logging.info('Format errors:')
        i = 0
        for tok in self._invalid_token:
            if tok.token.index is not None:
                logging.info('Id: ' + str(i) + ') ' + join(self.path_to_file, self.new_file_name) + ': ' +
                             str(tok.token.row) + '|' + str(tok.token.column) + ' Change from ' +
                             str(tok.old_value) + ' to ' + str(tok.new_value))

            i += 1
        logging.info('errors end\n')

    def setup_all_changes(self):
        for token in self._token:
            if token.type == Tokens.Identifier:
                value = self._symbol_table[token.index]
                for change in PHPFormatter.all_changes:
                    if change.format_type != 'var' and change.old_value == value:
                        self._symbol_table[token.index] = change.new_value
                        self._invalid_token.append(WrongToken(
                            token=token,
                            message=change.error_message,
                            format_type=change.format_type,
                            old_value=change.old_value,
                            new_value=change.new_value
                        ))
                        break



from .dict_token_types import Tokens


class Token:

    def __init__(self, token_type=None, row=None, column=None,  index=None, spec=None):
        self.type = token_type
        self.spec = spec
        self.index = index
        self.row = row
        self.column = column

    def set_invalid(self):
        self.index = self.row = self.column = self.spec = None
        self.type = Tokens.Invalid

    def is_fake(self):
        return self.type is None and self.spec is None and \
               self.index is None and self.row is None and self.column is None

    def __str__(self):
        ans = ''
        if self.type is not None:
            if self.spec is not None:
                ans += str(self.type).ljust(20)
            else:
                ans += str(self.type).ljust(40)
        if self.spec is not None:
            ans += (' ( ' + self.spec + ' ) ').ljust(20)
        if self.row is not None:
            ans += '|' + (str(self.row)).rjust(2) + '|'
        if self.column is not None:
            ans += (str(self.column)).ljust(2) + ('|').ljust(10)

        return ans


class WrongToken:

    def __init__(self, message=None, token=None, format_type=None, old_value=None, new_value=None):
        self.error_message = message
        self.token = token
        self.format_type = format_type
        self.old_value = old_value
        self.new_value = new_value

    def __str__(self):
        s = str(self.token) + ' | ' + self.error_message.ljust(50)
        if self.format_type is not None:
            s += ' | ' + self.format_type.ljust(12)
        return s



def is_number(c):
    return c in "01233456789"


def is_symbol(c):
    return 'a' <= c <= 'z' or 'A' <= c <= 'Z' or c == '_'


def is_dot(c):
    return c == '.'


def is_whitespace(c):
    return c == ' ' or c == '\t' or c == '\n'


def is_punctuation(c):
    return c in ";,[](){}"


def is_operation(c):
    return c in "+-*/%^|&=<>!?:~.\\"


def is_comment(c1, c2):
    return c1 == '/' and (c2 == '/' or c2 == '*')


def is_word(c):
    return is_symbol(c) or is_number(c)


def is_string(c):
    return c == '\'' or c == '"'


def is_correct_after_number(c):
    return is_punctuation(c) or is_whitespace(c) or is_operation(c)


def is_hex(c):
    return c in "0123456789AaBbCcDdEeFf"


def is_binary(c):
    return c in "01"


def is_octal(c):
    return c in "01234567"


def is_down_dash(c):
    return c == '_'


def is_e(c):
    return c in "Ee"


def is_nonnormal_start(state, check_func, type_letter):
    if state.line[state.column] == '0':
        if state.column + 1 < len(state.line) and state.line[state.column+1] in type_letter:
            if state.column + 2 < len(state.line):
                if check_func(state.line[state.column+2]):
                    return 1
            return 0
    return -1


def is_dollar(c):
    return c == '$'



import errno

from .dict_token_types import Tokens
from ._char_checks import *
from .token_classes import Token, WrongToken


class _CurrentLineState:
    all_lines = []

    def __init__(self):
        self.column = 0
        self.row = 0
        self.line = ""
        self.token = Token(token_type=Tokens.Invalid)
        self.multi_line_mode = False

    def copy(self, instance):
        self.row = instance.row
        self.column = instance.column
        self.line = instance.line


class Lexer:

    def __init__(self):
        self._symbol_table = []
        self._token = []
        self._invalid_token = []

    def process_php_file(self, path_to_file):
        try:
            f = open(path_to_file, "r")
            state = _CurrentLineState()
            all_lines = [line for line in f]
            _CurrentLineState.all_lines = all_lines

            while state.row < len(all_lines):
                state.line = all_lines[state.row]
                state.column = 0
                while state.column < len(state.line):
                    self._get_next_token(state)
                state.row += 1
            state.row -= 1
            state.column -= 1
            self._check_multi_line_mode(state)

        except IOError as x:
            if x.errno == errno.ENOENT:
                print(path_to_file, '- does not exist')
            elif x.errno == errno.EACCES:
                print(path_to_file, '- cannot be read')
            else:
                print(path_to_file, '- some other error')

    def get_tokens_list(self):
        return self._token

    def get_error_tokens_list(self):
        return self._invalid_token

    def get_symbol_table(self):
        return self._symbol_table

    def _check_multi_line_mode(self, state):
        if state.multi_line_mode:
            if state.token.type == Tokens.MultiLineComment:
                self._invalid_token.append(WrongToken(message="Multi line comment requires end",
                                                      token=state.token))
            if state.token.type == Tokens.StringLiteral and state.token.spec != '`':
                self._invalid_token.append(WrongToken(message="String literal requires the end",
                                                      token=state.token))
            if state.token.type == Tokens.StringLiteral and state.token.spec == '`':
                self._token.append(Token(token_type=Tokens.EndTemplateString,
                                         row=state.row,
                                         column=state.column))
                self._invalid_token.append(WrongToken(message="String literal requires the end",
                                                      token=self._token[-1]))

    def print_all(self):
        for tok in self._token:
            if tok.type != Tokens.Space and tok.type != Tokens.Tab:
                print(tok, end=' ')
                if tok.index is not None:
                    print('|' + self._symbol_table[tok.index] + '|')
                else:
                    print('')
            if tok.type == Tokens.Enter:
                print('')
        print('------------------------')
        for tok in self._invalid_token:
            print(tok, end=' ')
            if tok.token.index is not None:
                print('|' + self._symbol_table[tok.token.index] + '|')
            else:
                print('')

        print('------------------------')
        for i in range(len(self._symbol_table)):
            print(str(i) + ') ', '|' + self._symbol_table[i] + '|')

    def _get_next_token(self, state):
        if state.multi_line_mode:
            self._handle_multi_line_mode(state)
            return

        if len(state.line) <= state.column:
            return

        current_symbol = state.line[state.column]

        if is_whitespace(current_symbol):
            self._handle_whitespaces(state)
            return

        if is_string(current_symbol):
            self._handle_strings(state)
            return

        if state.column + 1 < len(state.line) and is_comment(current_symbol, state.line[state.column + 1]):
            self._handle_comment(state)
            return

        if is_number(current_symbol) or is_dot(current_symbol):
            self._handle_number(state)
            return

        if is_symbol(current_symbol):
            self._handle_word(state)
            return

        if is_dollar(current_symbol):
            self._handle_variable(state)
            return

        if is_operation(current_symbol):
            self._handle_operation(state)
            return

        if is_punctuation(current_symbol):
            self._handle_punctuation(state)
            return

        self._add_wrong_token(token_type=Tokens.Invalid,
                              row=state.row,
                              column=state.column,
                              lit_value=current_symbol,
                              error_message="Invalid symbol")
        state.column += 1

    def _handle_whitespaces(self, state):
        while state.column < len(state.line):
            current_symbol = state.line[state.column]
            if current_symbol == ' ':
                self._token.append(Token(token_type=Tokens.Space,
                                         row=state.row,
                                         column=state.column))
            elif current_symbol == '\t':
                self._token.append(Token(token_type=Tokens.Tab,
                                         row=state.row,
                                         column=state.column))
            elif current_symbol == '\n':
                self._token.append(Token(token_type=Tokens.Enter,
                                         row=state.row,
                                         column=state.column))
            else:
                break
            state.column += 1

    def _handle_punctuation(self, state):
        current_symbol = state.line[state.column]
        self._token.append(Token(token_type=Tokens.Punctuation,
                                 row=state.row,
                                 column=state.column,
                                 spec=current_symbol))
        state.column += 1

    def _add_wrong_token(self, token_type, row, column, lit_value, error_message):
        self._token.append(Token(token_type=token_type,
                                 row=row,
                                 column=column,
                                 index=len(self._symbol_table)))
        self._symbol_table.append(lit_value)
        self._invalid_token.append(WrongToken(message=error_message,
                                              token=self._token[-1]))

    def _add_correct_token(self, token_type, row, column, lit_value, token_spec=None):
        self._token.append(Token(token_type=token_type,
                                 row=row,
                                 column=column,
                                 index=len(self._symbol_table),
                                 spec=token_spec))
        self._symbol_table.append(lit_value)

    def _handle_nonnormal_number(self, state, check_func, type_number):
        start_pos = state.column
        new_number = state.line[state.column:state.column+2]
        state.column += 2
        last_bad = False
        while state.column < len(state.line):
            current_symbol = state.line[state.column]
            if check_func(current_symbol):
                new_number += current_symbol
                last_bad = False
            elif is_down_dash(current_symbol):
                last_bad = True
                new_number += current_symbol
            else:
                break

            state.column += 1

        if last_bad:
            self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                  row=state.row,
                                  column=start_pos,
                                  lit_value=new_number,
                                  error_message="Incorrect number ending")
            return

        if state.column >= len(state.line) or is_correct_after_number(state.line[state.column]):
            self._add_correct_token(token_type=Tokens.NumberLiteral,
                                    row=state.row,
                                    column=start_pos,
                                    lit_value=new_number,
                                    token_spec=Tokens.NumberLiteral.value[type_number])
            return

        if state.line[state.column] == 'n':
            # octal
            state.column += 1
            new_number += 'n'

            if type_number == 3 and new_number[1] == '0':
                self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                      row=state.row,
                                      column=start_pos,
                                      lit_value=new_number,
                                      error_message="Incorrect writing for big octal numbers")
                return

            elif state.column >= len(state.line) or is_correct_after_number(state.line[state.column]):
                self._add_correct_token(token_type=Tokens.NumberLiteral,
                                        row=state.row,
                                        column=start_pos,
                                        lit_value=new_number,
                                        token_spec=Tokens.NumberLiteral.value[type_number + 4])
                return

        self._add_wrong_token(token_type=Tokens.NumberLiteral,
                              row=state.row,
                              column=start_pos,
                              lit_value=new_number + state.line[state.column],
                              error_message="Incorrect number ending")
        state.column += 1

    def _check_nonnormal_numbers(self, state):
        hex_val = is_nonnormal_start(state, is_hex, 'Xx')
        oct_val = is_nonnormal_start(state, is_octal, '0oO')
        bin_val = is_nonnormal_start(state, is_binary, 'Bb')

        if hex_val == 1:
            self._handle_nonnormal_number(state, is_hex, 4)
            return True
        if oct_val == 1:
            self._handle_nonnormal_number(state, is_octal, 3)
            return True
        if bin_val == 1:
            self._handle_nonnormal_number(state, is_binary, 2)
            return True

        if hex_val*oct_val*bin_val == 0:
            ans = 'hex'
            if oct_val == 0:
                ans = 'oct'
            if bin_val == 0:
                ans = 'bin'
            self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                  row=state.row,
                                  column=state.column,
                                  lit_value=state.line[state.column:state.column+2],
                                  error_message="Invalid" + ans + "token")
            state.column += 2
            return True

        return False

    def _handle_number(self, state):
        if self._check_nonnormal_numbers(state):
            return

        current_symbol = state.line[state.column]

        if is_dot(current_symbol):
            if state.column + 1 >= len(state.line) or not is_number(state.line[state.column + 1]):
                if is_dot(state.line[state.column + 1]) and \
                          state.column + 2 < len(state.line) and is_dot(state.line[state.column + 2]):
                    self._token.append(Token(token_type=Tokens.Operators,
                                             spec='...',
                                             row=state.row,
                                             column=state.column))
                    state.column += 3
                else:
                    self._token.append(Token(token_type=Tokens.Punctuation,
                                             row=state.row,
                                             column=state.column,
                                             spec=current_symbol))
                    state.column += 1
                return

        if current_symbol == '0' and state.column + 1 < len(state.line) and \
                not is_correct_after_number(state.line[state.column + 1]) and \
                not is_dot(state.line[state.column + 1]):
            state.column += 1
            self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                  row=state.row,
                                  column=state.column-1,
                                  lit_value=current_symbol+state.line[state.column],
                                  error_message="Not allowed start from zero")
            state.column += 1
            return

        new_number = ""
        was_dot, was_e, was_sign, was_dash, last_dash = False, False, False, False, False
        start_pos = state.column
        unknown = False

        while state.column < len(state.line):
            current_symbol = state.line[state.column]
            if is_number(current_symbol):
                new_number += current_symbol
                last_dash = False

            elif is_dot(current_symbol):
                new_number += current_symbol
                if last_dash:
                    self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                          row=state.row,
                                          column=start_pos,
                                          lit_value=new_number,
                                          error_message="Unexpected symbol in the number end")
                    state.column += 1
                    return
                was_dot = True
                break

            elif is_e(current_symbol):
                new_number += current_symbol
                if last_dash:
                    self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                          row=state.row,
                                          column=start_pos,
                                          lit_value=new_number,
                                          error_message="Unexpected symbol in the number end")
                    state.column += 1
                    return
                was_e = True
                break

            elif is_down_dash(current_symbol):
                last_dash = True
                new_number += current_symbol

            else:
                unknown = True
                break

            state.column += 1

        if was_dot:
            state.column += 1
            if state.column < len(state.line) and \
                    (is_number(state.line[state.column]) or is_e(state.line[state.column])):
                while state.column < len(state.line):
                    current_symbol = state.line[state.column]
                    if is_number(current_symbol):
                        new_number += current_symbol
                        last_dash = False

                    elif is_e(current_symbol):
                        new_number += current_symbol
                        if last_dash:
                            self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                                  row=state.row,
                                                  column=start_pos,
                                                  lit_value=new_number,
                                                  error_message="Unexpected symbol in the number end")
                            state.column += 1
                            return

                        was_e = True
                        break
                    elif is_down_dash(current_symbol):
                        last_dash = True
                        new_number += current_symbol

                    else:
                        unknown = True
                        break
                    state.column += 1

        if was_e:
            state.column += 1
            if state.column < len(state.line) and \
                    (is_number(state.line[state.column]) or state.line[state.column] in '+-'):
                if state.line[state.column] in '+-':
                    new_number += state.line[state.column]
                    state.column += 1
                    if state.column >= len(state.line) or not is_number(state.line[state.column]):
                        new_number = new_number[:-1]
                        self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                              row=state.row,
                                              column=start_pos,
                                              lit_value=new_number,
                                              error_message="Unexpected symbol in the number end")

                        state.column -= 1
                        return

                while state.column < len(state.line):
                    current_symbol = state.line[state.column]
                    if is_number(current_symbol):
                        new_number += current_symbol
                        last_dash = False

                    elif is_down_dash(current_symbol):
                        last_dash = True
                        new_number += current_symbol

                    else:
                        unknown = True
                        break

                    state.column += 1

            else:
                self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                      row=state.row,
                                      column=start_pos,
                                      lit_value=new_number,
                                      error_message="Unexpected symbol in the number end")
                return

        if last_dash:
            self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                  row=state.row,
                                  column=start_pos,
                                  lit_value=new_number,
                                  error_message="Unexpected symbol in the number end")
            state.column += 1
            return

        if unknown:
            if is_correct_after_number(state.line[state.column]):
                cur_spec = Tokens.NumberLiteral.value[1]
                if was_dot or was_e:
                    cur_spec = Tokens.NumberLiteral.value[0]

                self._add_correct_token(token_type=Tokens.NumberLiteral,
                                        row=state.row,
                                        column=start_pos,
                                        lit_value=new_number,
                                        token_spec=cur_spec)

            elif state.line[state.column] == 'n':
                new_number += 'n'
                if was_dot or was_e:
                    self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                          row=state.row,
                                          column=start_pos,
                                          lit_value=new_number,
                                          error_message="Invalid biginteger literal")
                else:
                    self._add_correct_token(token_type=Tokens.NumberLiteral,
                                            row=state.row,
                                            column=start_pos,
                                            lit_value=new_number,
                                            token_spec=Tokens.NumberLiteral.value[5])
                state.column += 1
            else:
                self._add_wrong_token(token_type=Tokens.NumberLiteral,
                                      row=state.row,
                                      column=start_pos,
                                      lit_value=new_number+state.line[state.column],
                                      error_message="Unexpected symbol in the number end")
                state.column += 1
        else:
            cur_spec = Tokens.NumberLiteral.value[1]
            if was_dot or was_e:
                cur_spec = Tokens.NumberLiteral.value[0]

            self._add_correct_token(token_type=Tokens.NumberLiteral,
                                    row=state.row,
                                    column=start_pos,
                                    lit_value=new_number,
                                    token_spec=cur_spec)

    def _handle_word(self, state):
        word = ""
        start_pos = state.column
        while state.column < len(state.line):
            current_symbol = state.line[state.column]
            if not is_word(current_symbol):
                break
            word += current_symbol
            state.column += 1
        if word in Tokens.Keyword.value:
            self._token.append(Token(token_type=Tokens.Keyword,
                                     row=state.row,
                                     column=start_pos,
                                     spec=word))
            return
        if word in Tokens.ConstLiteral.value:
            self._token.append(Token(token_type=Tokens.ConstLiteral,
                                     row=state.row,
                                     column=start_pos,
                                     spec=word))
            return
        self._add_correct_token(token_type=Tokens.Identifier,
                                row=state.row,
                                column=start_pos,
                                lit_value=word)

    def _handle_variable(self, state):
        if state.column + 1 >= len(state.line) or not is_symbol(state.line[state.column + 1]):
            self._add_wrong_token(token_type=Tokens.Variable,
                                  row=state.row,
                                  column=state.column,
                                  lit_value='$',
                                  error_message="Not completed variable")
        word = "$"
        state.column += 1
        start_pos = state.column
        while state.column < len(state.line):
            current_symbol = state.line[state.column]
            if not is_word(current_symbol):
                break
            word += current_symbol
            state.column += 1
        if word in Tokens.Keyword.value:
            self._token.append(Token(token_type=Tokens.Keyword,
                                     row=state.row,
                                     column=start_pos,
                                     spec=word))
            return
        if word in Tokens.ConstLiteral.value:
            self._token.append(Token(token_type=Tokens.ConstLiteral,
                                     row=state.row,
                                     column=start_pos,
                                     spec=word))
            return
        self._add_correct_token(token_type=Tokens.Variable,
                                row=state.row,
                                column=start_pos,
                                lit_value=word)

    def _handle_operation(self, state):
        op = ""
        start_pos = state.column
        while state.column < len(state.line):
            temp_op = op + state.line[state.column]
            if temp_op in Tokens.Operators.value:
                op = temp_op
            else:
                break
            state.column += 1

        self._token.append(Token(token_type=Tokens.Operators,
                                 row=state.row,
                                 column=start_pos,
                                 spec=op))

    def _handle_single_line_comment(self, state):
        if state.line[-1] != '\n':
            comment = state.line[state.column:]
        else:
            comment = state.line[state.column:-1]
        self._add_correct_token(token_type=Tokens.SingleLineComment,
                                row=state.row,
                                column=state.column,
                                lit_value=comment)

        if state.line[-1] == '\n':
            state.column = len(state.line)-1
        else:
            state.column = len(state.line)

    def _handle_multi_line_comment(self, state):
        pos = state.line[state.column:].find("*/")
        if pos == -1:
            comment = state.line[state.column:]
            self._symbol_table[state.token.index] += comment
            state.column = len(state.line)
            state.multi_line_mode = True
        else:
            comment = state.line[state.column:state.column + pos + 2]
            self._symbol_table[state.token.index] += comment
            state.column += pos + 2
            state.multi_line_mode = False
            state.token = None

    def _init_multi_line_comment(self, state):
        self._add_correct_token(token_type=Tokens.MultiLineComment,
                                row=state.row,
                                column=state.column,
                                lit_value=state.line[state.column:state.column + 2])
        state.token = self._token[-1]
        state.multi_line_mode = True
        state.column += 2

    def _handle_comment(self, state):
        if state.line[state.column:state.column+2] == Tokens.SingleLineComment.value:
            self._handle_single_line_comment(state)
        else:
            self._init_multi_line_comment(state)

    def _handle_multi_line_mode(self, state):
        if state.token.type == Tokens.MultiLineComment:
            self._handle_multi_line_comment(state)
        elif state.token.type == Tokens.StringLiteral and state.token.spec == "'":
            self._handle_common_string(state)
        elif state.token.type == Tokens.StringLiteral and state.token.spec == '"':
            self._handle_template_string(state)
        else:
            state.multi_line_mode = False

    def _init_common_string(self, state):
        self._add_correct_token(token_type=Tokens.StringLiteral,
                                row=state.row,
                                column=state.column,
                                lit_value=state.line[state.column],
                                token_spec=state.line[state.column])
        state.token = self._token[-1]
        state.multi_line_mode = True
        state.column += 1

    def _handle_common_string(self, state):
        st = state.column
        ending = state.token.spec
        while st < len(state.line):
            temp_s = state.line[st:]
            pos = temp_s.find(ending)
            if pos == -1:
                string_lit = state.line[state.column:]
                self._symbol_table[state.token.index] += string_lit
                state.column = len(state.line)
                return
            else:
                st += pos
                if state.line[st - 1] == '\\':
                    slash_pos = st - 1
                    while slash_pos >= state.column and state.line[slash_pos] == '\\':
                        slash_pos -= 1
                    if (st - 1 - slash_pos) % 2 == 1:
                        st += 1
                        continue

                string_lit = state.line[state.column:st+1]
                self._symbol_table[state.token.index] += string_lit
                state.column = st + 1
                state.token = None
                state.multi_line_mode = False
                return

    def _init_template_string(self, state):
        self._token.append(Token(token_type=Tokens.StartTemplateString,
                                 row=state.row,
                                 column=state.column))
        self._add_correct_token(token_type=Tokens.StringLiteral,
                                row=state.row,
                                column=state.column,
                                lit_value="",
                                token_spec=state.line[state.column])
        state.token = self._token[-1]
        state.multi_line_mode = True
        state.column += 1

    def _get_next_token_interpolate(self, state):
        current_symbol = state.line[state.column]
        if not state.multi_line_mode and current_symbol == Tokens.InterpolationEnd.value:
            self._token.append(Token(token_type=Tokens.InterpolationEnd,
                                     row=state.row,
                                     column=state.column))
            return True
        else:
            self._get_next_token(state)
            return False

    def _interpolate(self, state):

        self._token.append(Token(token_type=Tokens.InterpolationStart,
                                 row=state.row,
                                 column=state.column))
        state.column += 2

        while state.row < len(state.all_lines):
            state.line = state.all_lines[state.row]
            while state.column < len(state.line):
                if self._get_next_token_interpolate(state):
                    return
            state.column = 0
            state.row += 1
        state.row -= 1
        state.column = len(state.all_lines[-1])-1
        self._check_multi_line_mode(state)

        self._token.append(Token(token_type=Tokens.InterpolationEnd,
                                 row=state.row,
                                 column=state.column))
        self._invalid_token.append(WrongToken(message="There was no end for interpolation",
                                              token=self._token[-1]))

    def _check_interpolation(self, state, start, end):
        line_slice = state.line[start:end]
        symbols = "${"
        pos = line_slice.find(symbols)
        if pos == -1:
            return False
        string_literal = state.line[state.column:state.column + pos]
        state.column += pos
        self._symbol_table[state.token.index] += string_literal
        new_state = _CurrentLineState()
        new_state.copy(state)
        self._interpolate(new_state)
        state.copy(new_state)

        self._add_correct_token(token_type=Tokens.StringLiteral,
                                row=state.row,
                                column=state.column,
                                lit_value='',
                                token_spec='"')
        state.token = self._token[-1]
        state.column += 1
        return True

    def _handle_template_string(self, state):
        st = state.column
        ending = state.token.spec
        while st < len(state.line):
            temp_s = state.line[st:]
            pos = temp_s.find(ending)
            if pos == -1:
                if self._check_interpolation(state, state.column, len(state.line)):
                    return

                string_lit = state.line[state.column:]
                self._symbol_table[state.token.index] += string_lit
                state.column = len(state.line)
                return
            else:
                st += pos
                if self._check_interpolation(state, state.column, st):
                    return
                if st-1 >= state.column and state.line[st - 1] == '\\':
                    slash_pos = st - 1
                    while slash_pos >= state.column and state.line[slash_pos] == '\\':
                        slash_pos -= 1
                    if (st - 1 - slash_pos) % 2 == 1:
                        st += 1
                        continue

                string_lit = state.line[state.column:st]
                self._symbol_table[state.token.index] += string_lit
                self._token.append(Token(token_type=Tokens.EndTemplateString,
                                         row=state.row,
                                         column=st))
                state.column = st + 1
                state.token = None
                state.multi_line_mode = False
                return

    def _handle_strings(self, state):
        if state.line[state.column] == "'":
            self._init_common_string(state)
        else:
            self._init_template_string(state)



from setuptools import setup, find_packages

classifiers = [
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
]

setup(
    name="symphony_of_tranquility_php_formatter",
    version="0.0.1",
    author="SymphonyOfTranquility",
    author_email="symphony.of.tranquility@gmail.com",
    description="A small php formatter",
    long_description="",
    url="",
    packages=find_packages(),
    entry_points={
        "console_scripts": ['PhpCCF=php_formatter.main:main']
        },
    python_requires='>=3.6',
)